//
//  glas.c
//  GLAS
//
//  Created by Stanko Juzbasic on 12/19/15.
//  Copyright (c) 2015 Stanko Juzbasic. All rights reserved.
//
//  This program combines features of CoreAudio, GLUT, GCD,TPCircularBuffer and CBuffer
//  based on POSIX threads
//  TPCircularBuffer (c) Michael Tyson, CBuffer (c) Willem Thiart
//  This is also the "dummy code" for developing parts of SculptView functionality
//
//  The code can run comparatively either full-POSIX mode or mixed pthread - libdispatch (GCD) mode.

#include <string.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "glas.h"



#define kInputBus 1
#define kOutputBus 0
#define kBufferLength 4096
#define twopi    6.283185307179586476925287
#define TWOPI    6.283185307179586476925287

#ifndef MIN
#define MIN(X, Y) (((X) < (Y)) ? (X) : (Y))
#endif

int msecs;
int par;

    Boolean oldstuff = false;

//_____________________________________________________________________________
void wip(void)
{
    printf("not implemented\n"); //work-in-progress
    return;
}
//_____________________________________________________________________________
/*------------------------------------------------------------*/
/* PURPOSE: Evaluate Bessel function of first kind and order  */
/*          0 at input x  -written by M.G.R. Vogelaar         */
/*          Copyright (c) 1998  Kapteyn Institute Groningen   */
/*------------------------------------------------------------*/
double bessj0( double x )
{
    double ax,z;
    double xx,y,ans,ans1,ans2;
    
    if ((ax=fabs(x)) < 8.0) {
        y=x*x;
        ans1=57568490574.0+y*(-13362590354.0+y*(651619640.7
                                                +y*(-11214424.18+y*(77392.33017+y*(-184.9052456)))));
        ans2=57568490411.0+y*(1029532985.0+y*(9494680.718
                                              +y*(59272.64853+y*(267.8532712+y*1.0))));
        ans=ans1/ans2;
    } else {
        z=8.0/ax;
        y=z*z;
        xx=ax-0.785398164;
        ans1=1.0+y*(-0.1098628627e-2+y*(0.2734510407e-4
                                        +y*(-0.2073370639e-5+y*0.2093887211e-6)));
        ans2 = -0.1562499995e-1+y*(0.1430488765e-3
                                   +y*(-0.6911147651e-5+y*(0.7621095161e-6
                                                           -y*0.934935152e-7)));
        ans=sqrt(0.636619772/ax)*(cos(xx)*ans1-z*sin(xx)*ans2);
    }
    return ans;
}
//_____________________________________________________________________________
void makeFreqEnv (int env, wipAudioController *gen)
{
//partialAmplitudes spectral evnelope (=reciprocal, quadratic, bessel, exponential)
    float temp = 0;
    if (env < 1) env = 1;
    if (env > 4) env = 4;
    int fe = env;
    for(int i = 0; i < gen->partials; i++ ){
        if (fe == 4) temp += fabs((float)bessj0((double)i));
        if (fe == 3) temp += exp(-.5*i);
        if (fe == 2) temp += 1./((float)(i*i + 2*i + 1));
        if (fe == 1) temp += 1./(float)(i+1);
    }
    temp = 1./temp;
    if(debug) printf("temp = %f\n", temp);
    
    for(int i = 0 ; i < gen->partials; i++){
        if (fe == 1) gen->a[i] = .3*temp/(float)(i+1);
        if (fe == 2) gen->a[i] = .3*temp/((float)(i*i + 2*i + 1));
        if (fe == 3) gen->a[i] = .3*temp*exp(-.5*i);
        if (fe == 4) gen->a[i] = .02*(float)(bessj0((double)(i-2))*bessj0((double)(i-2)))/temp;
    }
    gen->freqEnv = env;
    
}
//_____________________________________________________________________________
int getCPUnum (void)
{
    int numCPU;
    int mib[4];
    size_t len = sizeof(numCPU);
    
    /* set the mib for hw.ncpu */
    mib[0] = CTL_HW;
    mib[1] = HW_AVAILCPU;  // alternatively, try HW_NCPU;
    
    /* get the number of CPUs from the system */
    sysctl(mib, 2, &numCPU, &len, NULL, 0);
    
    //printf("Number of CPUs: %d\n", numCPU);
    return numCPU;   
}
//_____________________________________________________________________________
Float32* makeSin (int rate)
{
    Sin = calloc(rate, sizeof(Float32));
    for(int i = 0; i < rate; i++)
        Sin[i] = sinf(TWOPI*i/(Float32)rate);
    return Sin;
}
//_____________________________________________________________________________
float note2time(float note, int low, int hi, float start, float len)
{
    float t = 0.;
    float lim = len - (float)(512/R);
    float range = (float)(hi-low);
    
    if(note < (float)low)t=start;
    if(note >= (float)hi)t=lim;
    else t = start + (lim - start)*(note - low)/range;
    
    return t;
}
//_____________________________________________________________________________
int note2x(float note, int low, int hi, int width)
{
    //int x = 0;
    if(note < (float)low) return 0;
    if(note > (float)hi)  return width;
    float range = (float)(hi-low);
    int x = (int)((note - low) * width/range);
    //printf("---->\tx = %d\n",x);
    return x;
}
//_____________________________________________________________________________
void x2X(int x)
{
Xf = 1047.*x/(Float32)gCamera.screenWidth;
    if( x <= 0 )Xpos = 0;

    else if( x >= gCamera.screenWidth )Xpos = gCamera.screenWidth -1;

    else Xpos = x;
    
    Yf = 200.;

    return;
}
//_____________________________________________________________________________
void note2X(int note, int low, int hi, int width)
{
    int x ;
    if(note < low) x = 0;
    if(note >  hi) x = width;
    float range = (float)(hi-low);
    x = (int)((note - low) * width/range);
    Xf = 1047.*x/(Float32)gCamera.screenWidth;
    if( x <= 0 )Xpos = 0;
        
    else if( x >= gCamera.screenWidth )Xpos = gCamera.screenWidth -1;
            
    else Xpos = x;
                
    Yf = 200.;
return;
}
//_____________________________________________________________________________
void pitchf2X(float pitch, int low, int hi, int width)
{
    int x ;
    if(pitch < (float)low) x = 0;
    if(pitch >  (float)hi) x = width;
    float range = (float)(hi-low);
    x = (int)((pitch - low) * width/range);
    Xf = 1047.*x/(Float32)gCamera.screenWidth;
    if( x <= 0 )Xpos = 0;
    
    else if( x >= gCamera.screenWidth )Xpos = gCamera.screenWidth -1;
    
    else Xpos = x;
    
    Yf = 200.;
    return;
}
//_____________________________________________________________________________
void pitchf2XY(float pitch, float vel, int low, int hi, int width)
{
    int x ;
    if(pitch < (float)low) x = 0;
    if(pitch >  (float)hi) x = width;
    float range = (float)(hi-low);
    x = (int)((pitch - low) * width/range);
    Xf = 1047.*x/(Float32)gCamera.screenWidth;
    if( x <= 0 )Xpos = 0;
    
    else if( x >= gCamera.screenWidth )Xpos = gCamera.screenWidth -1;
    
    else Xpos = x;
    
    Yf = 200.;
      // Yf = (gCamera.screenHeight - y)/(Float32)gCamera.screenHeight;
    
    Yf = Yc = vel/127.;
    
    //float y = (vel/127.)*(Float32)gCamera.screenHeight;
    
    if(debug)printf("Yf = %f\n", Yf);
    return;
}

//_____________________________________________________________________________
void updateGen (void)
{
    gen.f[0] = updateFrequency(Xf/524.);
    if(debug)printf("gen.f = %f\n", gen.f[0]);
    for(int i = 1; i < gen.partials; i++)
        gen.f[i] = gen.f[0]*i;
    
    glutPostRedisplay();
}
//_____________________________________________________________________________
void Idlef(void)
{
    //glutPostRedisplay();
    return;
}
//_____________________________________________________________________________
void update (int val)
{
    if(val)glutPostRedisplay();
    if(debug)printf("*");
    return;
}

//_____________________________________________________________________________
void timerCB(int par)
{
    if(midi){
    update(1);
    glutTimerFunc(FAST, timerCB, FAST); // for next timer event
    }else {
    /*
    glutTimerFunc(SLOW, timerCB, SLOW);
    if(debug)printf(".");
     */
    tCallbackReg = GL_FALSE;
    }
}

//_____________________________________________________________________________

#pragma mark - Audio -
//_____________________________________________________________________________
int fbplay0 (long size)
{
    wip();return (int)size;
}
//_____________________________________________________________________________
int fbstop0 (void)
{
    wip();return 0;
}
//_____________________________________________________________________________
// generic result checker
//_____________________________________________________________________________
static int checkResultLite(OSStatus error, const char *operation)
{
	if (error == noErr) return 0;
	
	char str[20];
    
	// see if it appears to be a 4-char-code
	*(UInt32 *)(str + 1) = CFSwapInt32HostToBig(error);
	if (isprint(str[1]) && isprint(str[2]) && isprint(str[3]) && isprint(str[4])) {
		str[0] = str[5] = '\'';
		str[6] = '\0';
	} else
		// no, format as an integer
		sprintf(str, "%d", (int)error);
    return 1;
}
//_____________________________________________________________________________
// audio setup & initialization
//_____________________________________________________________________________
void createToneUnit (wipAudioController *gen)
{
    UInt32 size = sizeof(Float64);
    //UInt32 enableInput = 1;
    
    
	// Configure the search parameters to find the default I/O unit
	// kAudioUnitSubType_DefaultOutput on Mac OS X for output only, but
	// kAudioUnitSubType_HALOutput on Mac OS X can also capture input signal!
	AudioComponentDescription audioCompDesc;
	audioCompDesc.componentType = kAudioUnitType_Output;
	audioCompDesc.componentSubType = kAudioUnitSubType_HALOutput;
	audioCompDesc.componentManufacturer = kAudioUnitManufacturer_Apple;
	audioCompDesc.componentFlags = 0;
	audioCompDesc.componentFlagsMask = 0;
    
	
	// Check for the described input-enableable output unit
	AudioComponent halOutput = AudioComponentFindNext(NULL, &audioCompDesc);
    if(halOutput == NULL) printf("Couldn't find unit of subType HAL output\n");
//	NSAssert(audioRenderCallback, @"Couldn't find unit of subType HAL output");
	
	// Create a new unit based on this that we'll use for I/O
    checkResultLite(AudioComponentInstanceNew(halOutput, &(gen->toneUnit)),
                    "Couldn't get audio unit instance");
    
    //enable input
    /*
    checkResultLite(AudioUnitSetProperty (gen->toneUnit,
                                          kAudioOutputUnitProperty_EnableIO,
                                          kAudioUnitScope_Output,
                                          kOutputBus,
                                          &enableInput,
                                          sizeof(enableInput)),
                    "Couldn't enable input");
    */
    
    //get sample rate
    checkResultLite(AudioUnitGetProperty(gen->toneUnit,
                                         kAudioUnitProperty_SampleRate,
                                         kAudioUnitScope_Output,
                                         kOutputBus,
                                         &R,
                                         &size),
                    "Couldn't get sample rate");
    
	// Set tone rendering function on the unit
	AURenderCallbackStruct input;
	input.inputProc = audioRenderCallback;
	input.inputProcRefCon = gen;
    
    
    checkResultLite( AudioUnitSetProperty(gen->toneUnit,
                                          kAudioUnitProperty_SetRenderCallback,
                                          kAudioUnitScope_Input,
                                          kOutputBus,
                                          &input,
                                          sizeof(input)),
                    "Error setting input callback: ");
    
    UInt32 flag = 0;
    checkResultLite( AudioUnitSetProperty(gen->toneUnit,
                                          kAudioUnitProperty_ShouldAllocateBuffer,
                                          kAudioUnitScope_Output,
                                          kInputBus,
                                          &flag,
                                          sizeof(flag)),
                    "Error unsetting buffer:");
	
	// Set streamFormat to 32 bit, single channel, floating point, linear PCM
	const int four_bytes_per_float = 4;
	const int eight_bits_per_byte = 8;
	gen->streamFormat.mSampleRate = R;
	gen->streamFormat.mFormatID = kAudioFormatLinearPCM;
	gen->streamFormat.mFormatFlags =
    kAudioFormatFlagsNativeFloatPacked | kAudioFormatFlagIsNonInterleaved;
	gen->streamFormat.mBytesPerPacket = four_bytes_per_float;
	gen->streamFormat.mFramesPerPacket = 1;
	gen->streamFormat.mBytesPerFrame = four_bytes_per_float;
	gen->streamFormat.mChannelsPerFrame = 1;
	gen->streamFormat.mBitsPerChannel = four_bytes_per_float * eight_bits_per_byte;
    
    //set streamFormat
	checkResultLite(AudioUnitSetProperty (gen->toneUnit,
                                          kAudioUnitProperty_StreamFormat,
                                          kAudioUnitScope_Input,
                                          kOutputBus,
                                          &(gen->streamFormat),
                                          sizeof(AudioStreamBasicDescription)),
                    "Couldn't set StreamFormat on input scope / kOutputBus");

    //set global variables
    if(gen->f == NULL)gen->f = calloc(MAXP,sizeof(Float32));
    if(gen->prevf == NULL)gen->prevf = calloc(MAXP,sizeof(Float32));
    if(gen->phase == NULL)gen->phase = calloc(MAXP,sizeof(Float32));
    if(gen-> previous_phase_increment == NULL)
        gen->previous_phase_increment  = calloc(MAXP,sizeof(Float32));
    if(gen->a == NULL)gen->a = calloc(MAXP,sizeof(Float32));
    if(gen->preva == NULL) gen->preva = calloc(MAXP,sizeof(Float32));
    if(gen->arg   == NULL) gen->arg   = calloc(MAXP,sizeof(Float32));
    if(gen->previous_arg == NULL) gen->previous_arg = calloc(MAXP,sizeof(Float32));
    if(gen->val   == NULL) gen->val   = calloc(MAXP,sizeof(Float32));
    
    gen->sampleRate = R;
    
    //printf("f0 = %f\n",f0);
    
    for(int i = 0; i < partials; i++){
        gen->f[i] = f0*(i+1.) ;
        //printf("f[%d] = %f\n",i,gen->f[i]);
        gen->prevf[i] = gen->f[i];
        gen->phase[i] = 0.;
        gen->previous_phase_increment[i] = 0.;
        gen->arg[i] = 0;
        gen->previous_arg[i] = 0;
        gen->val[i] = 0.;
    }

    makeFreqEnv(4, gen);
    
    gen->state = idle;
    gen->playing = false;
    gen->wtable  = true;
    
    checkResultLite(AudioUnitInitialize(gen->toneUnit),
                    "Error initializing unit:");
    checkResultLite(AudioOutputUnitStart(gen->toneUnit),
                    "Error starting unit: ");
}
//_________________________________________________________________________
void createMIDI (wipAudioController *gen)
{
    //create MIDI input and client - - - - - - - - - - -
    gen->midiclient = 0;
    
    checkResultLite(MIDIClientCreate(CFSTR("MIDI_Client"),
                                NULL,
                                /*midiClientNotifyRefCon*/NULL,
                                &gen->midiclient),
               "MIDI Client Create Error\nRelaunch Application:");
    
    checkResultLite(MIDIInputPortCreate(gen->midiclient,
                                   CFSTR("MIDI_Input"),
                                   midiReadProc,
                                   NULL,
                                   &gen->midiin),
               "MIDI Port Create Error\nRelaunch Application:");
    
    //connect MIDI - - - - - - - - - - - - - - - - - - -
    
    ItemCount mSrcs = MIDIGetNumberOfSources();
    //NSLog(@"MIDI Sources: %ld\n", (long)mSrcs);
    printf("MIDI Sources: %ld\n", (long)mSrcs);
    ItemCount iSrc;
    
    for (iSrc=0; iSrc<mSrcs; iSrc++) {
        MIDIEndpointRef src = MIDIGetSource(iSrc);
        MIDIPortConnectSource(gen->midiin, src, NULL);
    }
    omni = TRUE;
    
}

//_____________________________________________________________________________
void disposeMIDI (void)
{
    OSStatus status;
    ItemCount nSrcs = MIDIGetNumberOfSources();
    ItemCount iSrc;
    
    if (nSrcs > 0)
        for (iSrc=0; iSrc<nSrcs; iSrc++) {
            MIDIEndpointRef src = MIDIGetSource(iSrc);
            status = MIDIPortDisconnectSource(gen.midiin, src);
        }
    MIDIPortDispose(gen.midiin);
    MIDIClientDispose(gen.midiclient);
    gen.midiclient = 0;
}


//_____________________________________________________________________________
//  initialize/uninitialize everything
//_____________________________________________________________________________
void setup_audio (void) {

    // Initialize buffer
    TPCircularBufferInit(&(gen.buffer), kBufferLength);
    // Initialize cbuffer
    gen.cbuffer = *cbuf_new(12);
    // Create the semaphore
    mSemaphore = dispatch_semaphore_create(0);
    // Setup audio
    createToneUnit(&gen);
    if(1){
        createMIDI(&gen);
        gen.midif = FALSE;
    }
    // Initialize wavetable
    Sin = makeSin((int)(gen.sampleRate));
    on = false;
    printf("\n\tAudio SET UP\n");
}
//_____________________________________________________________________________
void dealloc_audio (void) {
    // Release buffer resources
    TPCircularBufferCleanup(&(gen.buffer));
    // Clean up semaphore
    dispatch_release(mSemaphore);
    if(gen.toneUnit){
        AudioOutputUnitStop(gen.toneUnit);
		AudioUnitUninitialize(gen.toneUnit);
		AudioComponentInstanceDispose(gen.toneUnit);
        printf("\n\tAudio RELEASED\n");
    }
    if(pthread_cancel(workThreadFarm) == 0)printf("   ...success\n\n");
    //[super dealloc];
}
//_____________________________________________________________________________
Float32 updateFrequency(float sliderValue) //betw. 0. and 2.
{
    if(0)printf("Slider: %f\n",sliderValue);
    return (Float32)f0*(powf(2.,sliderValue));//betw. 1. and 4.
}
//_____________________________________________________________________________
void updateFrequencies(void) //betw. 0. and 2.
{
    //printf("Slider: %f\n",sliderValue);
    //return (Float32)f0*(powf(2.,sliderValue));//betw. 1. and 4.
    gen.prevf[0] = gen.f[0];
    for (int i = 1; i < partials; i++){
        gen.prevf[i] = gen.f[i];
        gen.f[i] = (i+1)*gen.f[0];
    }
    return;
}
//_____________________________________________________________________________
void midiCatcher(void)
{
    ;
}
//_____________________________________________________________________________
void readMidi (float xf)
{    
    gen.prevf[0] = gen.f[0];
    float sliderValue = xf/524.;
    gen.f[0] = (Float32)f0*(powf(2.,sliderValue));
    for (int i = 1; i < partials; i++){
        gen.prevf[i] = gen.f[i];
        gen.f[i] = (i+1)*gen.f[0];
    }
    return;
}
//_____________________________________________________________________________
//  dispatcher (producer queue)
//_____________________________________________________________________________
void dproducer (Boolean on, Boolean blah)
{
    if (on == true)
    {
        dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_HIGH, 0), ^{
            if((multithreading)||(NF))
                producerSum(on);
            else
                sproducer(on);
        });
    }
    return;
}
//_____________________________________________________________________________
// synthesis algorithm using sinf( )
//_____________________________________________________________________________
void *producerPT  (void *TN)
{
    Float32 phi;
    Float32 samples[4096];
    tData threadData;
    threadData = *((tData *)TN);
    int tid  = threadData.tid;
    int step = threadData.cpuCount;
    int *ret = calloc(1,sizeof(int));
    Float32 frames = (Float32)threadData.genD->frames;
    if(0){
        printf("Thread#: %d\n", tid);
        //printf("NumCPU = %d\n", step);
    }
    
    for(int k = tid; k < threadData.genD->numCand; k = k + step){
        if(0)printf("K = %d\t prev.amp = %f\n", k, threadData.genD->preva[k]);
        if(k >  threadData.genD->numCand)break;

        Float32 a, da;

        a = (k < threadData.genD->candidates ? threadData.genD->a[k] : 0.);

             da = (a - threadData.genD->preva[k])/frames;

        
        Float32 f = threadData.genD->f[k];        
        Float32 cycle=  threadData.genD->sampleRate/f;
        Float32 phase_increment = twopi/cycle;
        Float32 dphi = (phase_increment - threadData.genD->previous_phase_increment[k])/frames;
        phi  = threadData.genD->phase[k];
        
        for (UInt32 frame = 0; frame < threadData.genD->frames; frame++) {
            samples[frame]  += .3*sinf(phi);
            samples[frame]  *= (threadData.genD->preva[k] +frame*da);
            phi = phi + threadData.genD->previous_phase_increment[k] + frame*dphi;
            if (phi > twopi) phi -= twopi;
        }//end frame
        
        threadData.genD->phase[k] = phi;
        threadData.genD->previous_phase_increment[k] = phase_increment;
        threadData.genD->preva[k] = a;
        
        for (int i = 0; i < threadData.genD->frames; i++)fbuffW[tid][i] += samples[i];
    }//end k
    
    return (void*)ret;
}

//_____________________________________________________________________________
// synthesis algorithm using wavetable lookup
//_____________________________________________________________________________
void *producerPTN  (void *TN)
{
    float b;
    int     arg;
    float samples[4096];
    tData    threadData;
    threadData = *((tData *)TN);
    int tid  = threadData.tid;
    int step = threadData.cpuCount;
    int *ret = (int*)calloc(1,sizeof(int));
    int rate     =   (int)threadData.genD->sampleRate;
    Float32 frames = (Float32)threadData.genD->frames;
    
    if(0){
        printf("Thread#: %d\n", tid);
        printf("NumCPU = %d\n", step);
    }
    
    for(int k = tid; k < threadData.genD->numCand; k = k + step){
        if(debug)printf("K = %d\t prev.amp = %f\n", k, threadData.genD->preva[k]);
        if(k > threadData.genD->numCand)break;
        //threadData.genD->f[k] = threadData.genD->f[0]*(k+1);//updates freqs from GUI
 
        Float32 a, da;
        a = (k < threadData.genD->candidates ? threadData.genD->a[k] : 0.);
        da = (a - threadData.genD->preva[k])/frames;


        Float32 f = threadData.genD->f[k];
        Float32 deltaf = (f - threadData.genD->prevf[k]);
        Float32 df     = deltaf/frames;
        //int idf        = (int)df; //if(idf != 0)printf("IDF = %d  ", idf);
        //int dph;
        arg = threadData.genD->arg[k];
        //dph = (int)(threadData.genD->prevf[k]);
        Float32 blah = threadData.genD->prevf[k];
        for (UInt32 frame = 0; frame < threadData.genD->frames; frame++){
            blah += df;
            arg += blah;
            if(arg > rate)arg %= rate;
            b = .3*Sin[arg];
            b *= (threadData.genD->preva[k] + frame*da);//new
            samples[frame]  += b;
        }//end frame
        threadData.genD->prevf[k] = f; //threadData.genD->f[k];
        threadData.genD->preva[k] = (k <= threadData.genD->candidates ? a : 0.);
        threadData.genD->arg[k]   = arg;
        threadData.genD->preva[k] = a;
        for (int i = 0; i < threadData.genD->frames; i++)fbuffW[tid][i] += samples[i];
    }//end k
    
    return (void*)ret;
}

//_____________________________________________________________________________
//  p_threadable producer routine:
//      segments the code into 'num' parallel threads,
//      computed from the number of logical CPU cores
//_____________________________________________________________________________
void *producerPSum(void* pdata)
{
    int rc;
    wipAudioController*    data;
    data = ((wipAudioController *)pdata);
    int cpuNum = data->cpuNum;
    bool on = data->on;
    int *retSum = (int*)calloc(1,sizeof(int));
    pthread_t threads[cpuNum];
    tData thread_args[cpuNum];
    void* resulT;
    static float frames  [4096];
    int     bytesToCopy, bytes;
    int32_t availableBytes;
    Boolean wavet = data->wtable;
    
    while(on){

        if(data->state == stopping){;
            //printf("!\n");
        }
        if(data->state == idle){
            on = false;
            return (void*)retSum;
        }
        
        data->candidates = (UInt32)(data->partials*Yc);
        //update freqs.
        if(data->midif){
            data->f[0] = updateFrequency(Xf/524.);
            updateFrequencies();
        }
        Float32 damp = 1./(Float32)(gen.frames);
        bytesToCopy = data->frames*sizeof(float);
        memset(frames,  0, 4096*sizeof(float));
        availableBytes = 0;
        if( (fbuffW  = (float**)calloc(cpuNum + 1, sizeof(float*)))!= NULL)
            for (int i=0; i<cpuNum; ++i)
            {
                fbuffW[i] = (float*)calloc(data->frames, sizeof(float));
                thread_args[i].tid = i; //ord. number of thread
                thread_args[i].cpuCount = cpuNum; //counter increment step
                thread_args[i].genD = data;
                if (wavet)rc = pthread_create(&threads[i], NULL, producerPTN, (void *) &thread_args[i]);
                else rc = pthread_create(&threads[i], NULL, producerPT , (void *) &thread_args[i]);
                
                if(rc != 0)printf("rc = %s\n",strerror(rc));
            }
        
        for (int i=0; i<cpuNum; ++i) rc = pthread_join(threads[i], &resulT);
        
        for(UInt32 samp = 0; samp < data->frames; samp++)
            for(int i = 0; i < cpuNum; i++)
                frames[samp] += fbuffW[i][samp];
        
        switch (data->state) { //graceful suspension and termination...
            case idle:    memset(frames, 0, data->frames*sizeof(float)); break;
            case stopped:
                memset(frames, 0, data->frames*sizeof(float));
                data->state = idle; //on = FALSE;
                break;
            case starting:
                for(int i = 0; i < data->frames; i++)frames[i] *= (i*damp);
                data->state = active;
                break;
            case stopping:
                for(int i = 0; i < data->frames; i++)frames[i] *= (data->frames - i)*damp;
                data->state = stopped;
                //printf("?\n");
                break;
            default: break;
        }
        /*
        float *head = (float*)TPCircularBufferHead(&(data->buffer), &availableBytes);
        memcpy(head,(const void*)frames,MIN(bytesToCopy, availableBytes));//copies frames to head
        TPCircularBufferProduce(&(data->buffer),MIN(bytesToCopy,availableBytes));
        */
        bytes = bytesToCopy;
        int numBytes = cbuf_offer(&(data->cbuffer), (unsigned char*)frames, bytes);
        numBytes = 0;
        
        pthread_cond_wait (&cond, &mutex);
        
        for(int i = 0; i < cpuNum; i++)
            free(fbuffW[i]);
        if (fbuffW)free(fbuffW);
    } //end while(on)
    return (void*)retSum;
}

//___________________________________________________________________________________
// worker thread launcher - for all producer routines
//___________________________________________________________________________________
pthread_t pdproducer (bool on, bool wtable, wipAudioController* data)
{
    pthread_t sumThread;
    data->on = on;
    data->state = starting;
    int rcPsum;
    rcPsum = pthread_create(&sumThread, NULL, producerPSum, (void *) data);
    return sumThread;
}
//_____________________________________________________________________________
//  simplest producer routine
//_____________________________________________________________________________
void sproducer (Boolean on)
{
    int     bytesToCopy, bytes;
    /*
     Float32 samplesO[gen.frames];
     Float32 samplesE[gen.frames];
     Float32 frames [gen.frames];
     */
    static Float32 samplesO[4096];
    //static Float32 samplesE[4096];
    static Float32 framesO [4096];
    //static Float32 framesE [4096];
    static Float32 frames  [4096];
    Float32 phi;
    Float32 wait_interval;
    numCand = partials;

    
    while(on){
        wait_interval = 1.5*(gen.frames)/(gen.sampleRate);
        glutPostRedisplay();
        availableBytes = 0;
        
        candidates = (UInt32)(partials*Yc);
        
        //if((Xf > 1000.)||(Xf < 47.))candidates = 0;
        
        //update freqs.
        if(midi) {float blah = Xf;readMidi(blah); }
        
        memset(frames,   0, gen.frames*sizeof(Float32));
        memset(framesO,  0, gen.frames*sizeof(Float32));
        Float32 damp = 1./(Float32)(gen.frames);
        bytesToCopy = gen.frames*sizeof(Float32);
        //__Compute     partials_____________________________________________
        for(int k = 0; k < numCand; k = k+1){
            memset(samplesO, 0., gen.frames*sizeof(Float32));
          
            Float32 da;
            if ( k <= candidates )da = (gen.a[k] - gen.preva[k])/(Float32)(gen.frames);
            else da = -1*gen.preva[k]/(Float32)(gen.frames);
            Float32 cycle= gen.sampleRate/gen.f[k];
            Float32 phase_increment = twopi/cycle;
            Float32 dphi = (phase_increment - gen.previous_phase_increment[k])/(Float32)(gen.frames);
            phi  = gen.phase[k];
            
            for (int i = 0; i < gen.frames; i++) {
                samplesO[i]  += .3*sinf(phi);
                samplesO[i]  *= (gen.preva[k] +i*da);
                phi = phi + gen.previous_phase_increment[k] + i*dphi;
                if (phi > twopi) phi -= twopi;
            }
            
            gen.phase[k] = phi;
            gen.previous_phase_increment[k] = phase_increment;
            gen.preva[k] = k <= candidates ? gen.a[k] : 0.;
            for (int i = 0; i < gen.frames; i++)framesO[i] += samplesO[i];
        }//end 
        

        //__copy results_____________________________________________________
        for (int i = 0; i < gen.frames; i++)frames[i] = framesO[i];
        //for (int i = 0; i < gen.frames; i++)frames[i] = 0.;
        //___________________________________________________________________
        if(gen.state == idle)    memset(frames, 0., gen.frames*sizeof(Float32));
        if(gen.state == stopped) memset(frames, 0., gen.frames*sizeof(Float32));
        if(gen.state == starting){
            for(int i = 0; i < gen.frames; i++)frames[i] *= i*damp;
            gen.state = active;
        }
        if(gen.state == stopping){
            for(int i = 0; i < gen.frames; i++)
                frames[i] *= (gen.frames - i)*damp;
            gen.state = stopped;
        }

        bytes = bytesToCopy;

        int numBytes = cbuf_offer(&(gen.cbuffer), (unsigned char*)frames, bytes);
        numBytes = 0;
       
        
        // Wait for a condition
        pthread_cond_wait (&cond, &mutex);

        if(gen.state == stopped){gen.state = idle; on = false;}
    }
    if(debug)printf("RETURNING\n");
    //printf("%d\n",gen.state);
    return;
}
//_____________________________________________________________________________
//  updateable producer routine odd and even partials
//_____________________________________________________________________________
void producer (Boolean on)
{
    int     bytesToCopy, bytes;
    /*
    Float32 samplesO[gen.frames];
    Float32 samplesE[gen.frames];
    Float32 frames [gen.frames];
     */
    static Float32 samplesO[4096];
    static Float32 samplesE[4096];
    static Float32 framesO [4096];
    static Float32 framesE [4096];
    static Float32 frames  [4096];
    Float32 phi;
    Float32 wait_interval;
    numCand = partials;
    
    while(on){
        wait_interval = 1.5*(gen.frames)/(gen.sampleRate);
        availableBytes = 0;
        candidates = (UInt32)(partials*Yc);
        //update freqs.
        if(midi) {float blah = Xf;readMidi(blah);}
        
        
        //bytesToCopy = 0;
        memset(frames,   0, gen.frames*sizeof(Float32));
        memset(framesO,  0, gen.frames*sizeof(Float32));
        memset(framesE,  0, gen.frames*sizeof(Float32));
        Float32 damp = 1./(Float32)(gen.frames);
        bytesToCopy = gen.frames*sizeof(Float32);
        //__Compute odd partials_____________________________________________
        for(int k = 0; k < numCand; k = k+2){
        memset(samplesO, 0., gen.frames*sizeof(Float32));
        //gen.f[k] = gen.f[0]*(k+1); //updates freqs from GUI
            
        Float32 da;
        if ( k <= candidates )da = (gen.a[k] - gen.preva[k])/(Float32)(gen.frames);
        else da = -1*gen.preva[k]/(Float32)(gen.frames);
        Float32 cycle= gen.sampleRate/gen.f[k];
        Float32 phase_increment = twopi/cycle;
        Float32 dphi = (phase_increment - gen.previous_phase_increment[k])/(Float32)(gen.frames);
        phi  = gen.phase[k];
            
        for (int i = 0; i < gen.frames; i++) {
            samplesO[i]  += .3*sinf(phi);
            samplesO[i]  *= (gen.preva[k] +i*da);
            phi = phi + gen.previous_phase_increment[k] + i*dphi;
            if (phi > twopi) phi -= twopi;
            
        }
        
        gen.phase[k] = phi;
        gen.previous_phase_increment[k] = phase_increment;
        gen.preva[k] = k <= candidates ? gen.a[k] : 0.;
        for (int i = 0; i < gen.frames; i++)framesO[i] += samplesO[i];
        }//end odd partials
        
        //__Compute even partials_____________________________________________
        if(numCand > 1)
        for(int k = 1; k < numCand; k = k+2){
            memset(samplesE, 0., gen.frames*sizeof(Float32));
            //gen.f[k] = gen.f[0]*(k+1);//updates freqs from GUI
            /*
            if((Xf < 0)||(Xf > 1047))gen.a[k] = 0.;
            if(k >= numCand)gen.a[k] = 0.;
            else gen.a[k] = 1./(Float32)((k+1)*(k+1));
            */
            Float32 da;
            if ( k <= candidates )da = (gen.a[k] - gen.preva[k])/(Float32)(gen.frames);
            else da = -1*gen.preva[k]/(Float32)(gen.frames);
            
            Float32 cycle= gen.sampleRate/gen.f[k];
            Float32 phase_increment = twopi/cycle;
            Float32 dphi = (phase_increment - gen.previous_phase_increment[k])/(Float32)(gen.frames);
            phi  = gen.phase[k];
            
            for (int i = 0; i < gen.frames; i++) {
                samplesE[i]  += .3*sinf(phi);
                samplesE[i]  *= (gen.preva[k] +i*da);
                phi = phi + gen.previous_phase_increment[k] + i*dphi;
                if (phi > twopi) phi -= twopi;
                
            }
            
        gen.phase[k] = phi;
        gen.previous_phase_increment[k] = phase_increment;
        gen.preva[k] = k <= candidates ? gen.a[k] : 0.;

        for (int i = 0; i < gen.frames; i++)framesE[i] += samplesE[i];
        }//end even partials
        
        //__Sum odd and even results_________________________________________
        for (int i = 0; i < gen.frames; i++)frames[i] = framesO[i] + framesE[i];
        //___________________________________________________________________
        if(gen.state == idle)    memset(frames, 0., gen.frames*sizeof(Float32));
        if(gen.state == stopped) memset(frames, 0., gen.frames*sizeof(Float32));
        if(gen.state == starting){
            for(int i = 0; i < gen.frames; i++)frames[i] *= i*damp;
            gen.state = active;
        }
        if(gen.state == stopping){
            for(int i = 0; i < gen.frames; i++)frames[i] *= (gen.frames - i)*damp;
            gen.state = stopped;
        }
        bytes = bytesToCopy;
        /*
        //__Pass the summed result to circular buffer_________________________
        float *head = TPCircularBufferHead(&(gen.buffer), &availableBytes);

        //bytes = MIN(bytesToCopy, availableBytes);
        memcpy(head,(const void*)frames, bytes);//copies frames to head
        TPCircularBufferProduce(&(gen.buffer),bytes);
        
        */
        int numBytes = cbuf_offer(&(gen.cbuffer), (unsigned char*)frames, bytes);
        numBytes = 0;
        
        
        // Wait for a signal
        pthread_cond_wait (&cond, &mutex);
        if(gen.state == stopped){gen.state = idle; on = false;}
        if(midi) {float blah = Xf;readMidi(blah);}
    }
    if(debug)printf("RETURNING\n");
    //printf("%d\n",gen.state);
        return;
}
//_____________________________________________________________________________
//  threadable producer routines
//_____________________________________________________________________________
void producerSum(Boolean on)
{
    int rc;
    int num = getCPUnum();
    pthread_t threads[num];
    //int thread_args[num];
    data thread_args[num];
    void* resulT;
    static Float32 frames  [4096];
    Float32 wait_interval;
    int     bytesToCopy, bytes;
    Float32 floatmax;
    
    while(on){
        candidates = (UInt32)(partials*Yc);
        //update freqs.
        if(gen.midif){
            gen.f[0] = updateFrequency(Xf/524.);
            updateFrequencies();
        }
        floatmax = 0.;
        wait_interval = 1.3*(gen.frames)/(gen.sampleRate);
        Float32 damp = 1./(Float32)(gen.frames);
        bytesToCopy = gen.frames*sizeof(Float32);
        memset(frames,  0., 4096*sizeof(Float32));
        availableBytes = 0;
        
        fbuffW  = (Float32**)calloc(num + 1, sizeof(Float32*));
        
        for (int i=0; i<num; ++i)
        {
            fbuffW[i] = (Float32*)calloc(gen.frames, sizeof(Float32));
            thread_args[i].tid = i; //ord. number of thread
            thread_args[i].cpuCount = num; //counter increment step
            if (multithreading)rc = pthread_create(&threads[i], NULL, producerT , (void *) &thread_args[i]);
            if (NF) rc = pthread_create(&threads[i], NULL, producerTN, (void *) &thread_args[i]);
        }
        
        for (int i=0; i<num; ++i) {
            rc = pthread_join(threads[i], &resulT);
            //thread_sum += *(int*)result;
        }

        
        for(UInt32 samp = 0; samp < gen.frames; samp++)
        {
        for(int i = 0; i < num; i++)
                frames[samp] += fbuffW[i][samp];
        }//end samp

        
        if(gen.state == idle)    memset(frames, 0., gen.frames*sizeof(Float32));
        if(gen.state == stopped) memset(frames, 0., gen.frames*sizeof(Float32));
        if(gen.state == starting){
            for(int i = 0; i < gen.frames; i++)frames[i] *= i*damp;
            gen.state = active;
        }
        if(gen.state == stopping){
            for(int i = 0; i < gen.frames; i++)frames[i] *= (gen.frames - i)*damp;
            gen.state = stopped;
        }
        
        //float *head = TPCircularBufferHead(&(gen.buffer), &availableBytes);
        //memcpy(head,(const void*)frames,MIN(bytesToCopy, availableBytes));//copies frames to head
        //TPCircularBufferProduce(&(gen.buffer),MIN(bytesToCopy,availableBytes));
        
        bytes = bytesToCopy;
        int numBytes = cbuf_offer(&(gen.cbuffer), (unsigned char*)frames, bytes);
        numBytes = 0;
    
    pthread_cond_wait (&cond, &mutex);
    if(gen.state == stopped){gen.state = idle; on = false;}
    /**/
    for(int i = 0; i <= num; i++)
            free(fbuffW[i]);
        free(fbuffW);
        
    } //end while(on)
    if(debug)printf("RETURNING\n");

return;
}
//_____________________________________________________________________________
// mathematically correct algorithm using sinf() and values
//_____________________________________________________________________________
void *producerT  (void *T)
{
    Float32 phi;
    Float32 samples[4096];
    data threadData;
    threadData = *((data *)T);
    int tid  = threadData.tid;
    int step = threadData.cpuCount;
    int *ret = calloc(1,sizeof(int));
    if(0){
        printf("Thread#: %d\n", tid);
        //printf("NumCPU = %d\n", step);
    }

    for(int k = tid; k < gen.numCand; k = k + step){
        if(0)printf("K = %d\t prev.amp = %f\n", k, gen.preva[k]);
        if(k > gen.numCand)break;

        Float32 f = gen.f[k];

        Float32 da;
        if ( k <= gen.candidates )da = (gen.a[k] - gen.preva[k])/(Float32)(gen.frames);
        else da = -1*gen.preva[k]/(Float32)(gen.frames);
        
        Float32 cycle= gen.sampleRate/f;
        Float32 phase_increment = twopi/cycle;
        Float32 dphi = (phase_increment - gen.previous_phase_increment[k])/(Float32)(gen.frames);
        phi  = gen.phase[k];
        
        for (UInt32 i = 0; i < gen.frames; i++) {
            samples[i]  += .3*sinf(phi);
            samples[i]  *= (gen.preva[k] +i*da);
            phi = phi + gen.previous_phase_increment[k] + i*dphi;
            if (phi > twopi) phi -= twopi;            
        }//end i
        
        gen.phase[k] = phi;
        gen.previous_phase_increment[k] = phase_increment;
        gen.preva[k] = k <= gen.candidates ? gen.a[k] : 0.;
        
        for (int i = 0; i < gen.frames; i++)fbuffW[tid][i] += samples[i];
    }//end k
    
    return (void*)ret;
}
//_____________________________________________________________________________
// faster algorithm using wavetables and pointer arithmetics
//_____________________________________________________________________________
void *producerTN  (void *TN)
{
    Float32 b;
    int     arg;    
    Float32 samples[4096];
    data    threadData;
    threadData = *((data *)TN);
    int tid  = threadData.tid;
    int step = threadData.cpuCount;
    int *ret = (int*)calloc(1,sizeof(int));
    int rate       = (int) gen.sampleRate;
    Float32 frames = (Float32)gen.frames;
    
    if(0){
        printf("Thread#: %d\n", tid);
        printf("NumCPU = %d\n", step);
    }
    
    for(int k = tid; k < gen.numCand; k = k + step){
        if(0)printf("K = %d\t prev.amp = %f\n", k, gen.preva[k]);
        if(k > gen.numCand)break;
        

        Float32 a, da;
        a = gen.a[k];
        if ( k <= gen.candidates )
             da = (a - gen.preva[k])/frames;
        else da = -1*gen.preva[k]/frames;
        
        Float32 f = gen.f[k];
        Float32 deltaf = (f - gen.prevf[k]);
        Float32 df     = deltaf/frames;
        int idf        = (int)df; //if(idf != 0)printf("#IDF = %d  ", idf);
        int dph;       
        arg = gen.arg[k];
        dph = (int)(gen.prevf[k]);
        
        for (UInt32 frame = 0; frame < gen.frames; frame++){
            dph += idf;
            arg += dph;
            if(arg > rate)arg %= rate;
            b = .2*Sin[arg];
            samples[frame]  += b;
            samples[frame]  *= (gen.preva[k] +frame*da);
        }//end frame
        gen.prevf[k] = f;
        gen.preva[k] = k <= gen.candidates ? gen.a[k] : 0.;
        gen.arg[k]   = arg;
        
        for (int i = 0; i < gen.frames; i++)fbuffW[tid][i] += samples[i];
    }//end k
    
    return (void*)ret;
}
//_____________________________________________________________________________
// midiReadProc()
// monophonic controller with last held note priority.
// depending on manufacturers, we may run into any of the two "note-off" scenarios
//      true note-off-message and zero-velovity-note-on-message
//      we have to check against both of them
//_____________________________________________________________________________
void midiReadProc(const MIDIPacketList *packetList,
                  void* readProcRefCon,
                  void* srcConnRefCon)
{
    static Byte note; //has to be static, else gets deallocated by pitch mend data
    static Byte velocity;
    //int sourceUniqueID = *(int*)srcConnRefCon;
    MIDIPacket *packet = (MIDIPacket*)packetList->packet;

    //upper and lower MIDI note limit
    int low= 48;
    int hi = 72;
    if(midi)numCand = partials;    
    int count = packetList->numPackets;
    //outermost loop
    for (int k=0; k<count; k++) {
        //printPacketInfo(packet);
        Byte midiStatus = packet->data[0];
        Byte midiChannel= midiStatus & 0x0F;
        Byte midiCommand = midiStatus >> 4;
        
        //if midi
        if(midi){
            //if ((midiCommand == 0x08) ||(midiCommand == 0x09))
            //get MIDI channel
            if(midiCommand != 0xF)
            {
                //debug
                if(debug)printf("Channel: %x, Message: %x\n", midiChannel, midiCommand);
                if (!omni)
                    if(midiChannel != channel)
                        goto next_packet;
                if(midiCommand == '\v'){
                    printf("Ped: 0x0%x\n",midiCommand);
                }
                
                if(midiCommand == 0x09)
                {
                    note = packet->data[1] & 0x7F;
                    velocity = packet->data[2] & 0x7F;
                    
                    //hacked note-off!!!
                    if (velocity == 0x0){
                        //'note' is the note released!!
                        if( hold > 0){
                            //find which held note has been released!
                            for(int i = 1; i <= hold; i++)
                                //get the note and its ord. number
                                if(note == n[i]){
                                    //last note released
                                    if (i == hold) {
                                        goto h;
                                    }
                                    //any other note released ?
                                    //compact 'n' array, keep last note's highest position
                                    for (int j = i+1; j <= hold; j++)
                                        n[j-1] = n[j];
                                }
                        h:
                            //
                            if(hold > 0){
                                note = n[hold-1];
                                //printf("1Note: %d\n", note);
                                }//1
                            if(hold > 1)//prevents note = 0
                                pitchf = (float)note + bendf;

                            pitchf2X(pitchf, low, hi, gCamera.screenWidth);
                            prev_hold = hold;//089
                            hold--;//this enables last note to be heard
                            //
                        }
                        //note on!!!
                    }else{
                        if(debug)printf("Vel = %d\n", (int)velocity);
                        prev_hold = hold;//089
                        hold++;
                        //printf("2Note: %d\n", note);//2
                        ////
                        pitchf = (float)note + bendf;

                        pitchf2XY(pitchf, (float)velocity, low, hi, gCamera.screenWidth);
                        //printf("Xf = %f\n", Xf);
                        
                        //if(midi)
                        if((hold == 1)&&(prev_hold == 0)){
                            if(gen.state == idle){
                            on = true; gen.playing = TRUE;gen.state = starting;
   
                            dproducer(true, false);
                            }else if (gen.state == stopped)gen.state = starting;
                            else if (gen.state == stopping)gen.state = active;
                            
                            //printf("||||||||||||||||||\n");
                        }
                    n[hold] = note;
                    }
                    //printf("HOLD = %d\n",hold);
                    
                    if(hold ==  0){
                        if(gen.state == active)gen.state = stopping;
                        //on = false;
                        //printf("*****************\n");
                    }
                }
                int prevn;
                //proper note-off
                if(midiCommand == 0x08)
                {
                    note = packet->data[1] & 0x7F;
                    if( hold > 0){
                        //find which note has been released!
                        for(int i = 1; i <= hold; i++)
                            if(note == n[i]){
                                if (i == hold) {
                                    goto hh;
                                }
                                for (int j = i+1; j <= hold; j++)
                                    n[j-1] = n[j];
                            }
                    hh:
                        //
                        
                        prevn = (int)note;
                        if(hold > 0){
                            note = n[hold-1];//last held note
                             printf("1Note: %d\n", note);
                        }//1
                        if(hold > 1)//prevents note = 0
                            pitchf = (float)note + bendf;
                        pitchf2X(pitchf, low, hi, gCamera.screenWidth);
                        if (hold <= 1){
                            amplitudef = 0.;
                        }
                        prev_hold = hold;//089
                        hold--;//this enables last note to be heard
                        printf("OFF: %d\tHELD:%d\thold = %d\n",prevn,note,hold);
                        //
                    }
                    if(hold ==  0){
                        if(gen.state == active)gen.state = stopping;
                        //on = false;
                        //printf("*****************\n");
                    }
                }//end proper note-off
                //0xB -control change: "1011"
                if(midiCommand == 0xB){
                    Byte ctrl = packet->data[1] & 0x7F;
                    //printf("Ctrl = 0x%x\n", ctrl);
                    if (ctrl == 0x40){ //sustain pedal
                        Byte val  = packet->data[2] & 0x7F;
                        //printf("Ctrl = 0x%x\n", val);
                        if ( val == 0x0 ) if(gen.state == active){hold = 0;gen.state = stopping;}
                    }
                    if (ctrl == (0x01 & 0x7F)){          //C01 -> mod. depth
                        Byte val  = packet->data[2] & 0x7F;
                        bmi  = val;
                        mmif = (Float32)val/127.   ;
                        changed = TRUE;
                        if(debug)printf("mod:%f\n", mmif);
                    }
                    if (ctrl == (0x0B & 0x7F)){         //C11 -> expression
                        Byte val  = packet->data[2] & 0x7F;
                        bmi = val;
                        mxpf = 3*(Float32)val/127. + 1 ;
                        changex = TRUE;
                        if(debug)printf("xpr:%f\n",mxpf);
                    }
                }
                //pitch bend: 0xE
                if(midiCommand == 0xE){
                    //n[hold] = note;
                    Byte valm = packet->data[2] & 0x7F;
                    Byte vall = packet->data[1] & 0x7F;
                    bendf = 128*(Float32)valm + (Float32)vall - 8192.;
                    bendf = bendf/8192.; //betw. -1. and 1.
                    pitchf = (float)note + bendf;
                    printf("pbend = %f\n", bendf);
                    printf("3Note: %d\n", note);//3
                    //// must be here, else won't get noticed

                    pitchf2XY(pitchf, (float)velocity, low, hi, gCamera.screenWidth);
                }
            }
        }//end if midi
    next_packet:
        packet = MIDIPacketNext(packet);
        if (hold == 0) {
        prev_hold = 0;//089
        //printf("4Note: %d\n", note);//4
        //pitchf2XY(pitchf, (float)velocity, low, hi, gCamera.screenWidth);
        }
        if (!gen.playing)hold = prev_hold = 0;
    }//end for (k = 0; ...;...)
}
//_____________________________________________________________________________
//  render callback (consumer thread)
//_____________________________________________________________________________
static OSStatus audioRenderCallback(void *inRefCon,
                                    AudioUnitRenderActionFlags *ioActionFlags,
                                    const AudioTimeStamp *inTimeStamp,
                                    UInt32 inBusNumber,
                                    UInt32 inNumberFrames,
                                    AudioBufferList *ioData) {
    
    wipAudioController *THIS = (wipAudioController *)inRefCon;
 
    Float32 rad = (Float32)(twopi/inNumberFrames);
    // An event happens in the render thread- signal whoever is waiting
    if (THIS->state == active){
        pthread_mutex_lock(&mutex);
    }
    // Mono tone generator: we only need one buffer
    const int channel = 0;
    Float32* targetBuffer = (Float32 *)ioData->mBuffers[channel].mData;
    memset(targetBuffer,0.,inNumberFrames*sizeof(Float32));
    
    // Pull samples from circular buffer
   /* 
    int32_t availableBytes;
    Float32 *buffer = TPCircularBufferTail(&THIS->buffer, &availableBytes);
    
    //copy circularBuffer content to target buffer
    
    int bytesToCopy = ioData->mBuffers[0].mDataByteSize;
    if(0)printf("Frames: %u\tBytes: %d\n", inNumberFrames, bytesToCopy);
    memcpy(targetBuffer, (const void*)buffer, MIN(bytesToCopy, availableBytes));
    TPCircularBufferConsume(&THIS->buffer, availableBytes);
    */
    /*
    if(THIS->state == idle)
        memset(targetBuffer,0,inNumberFrames*sizeof(Float32));
    else{
   */     
        int used = cbuf_usedspace(&(THIS->cbuffer));
        unsigned char *ptr0 = cbuf_peek(&(THIS->cbuffer));
        if(ptr0 != NULL)memcpy(targetBuffer, (const void*)ptr0, used);
        cbuf_poll(&(THIS->cbuffer), used);
        //if(debug)NSLog(@"usedBytes= %d\n",used);
/*
    }
*/    
    THIS->frames = inNumberFrames;
    
    if(debug)for(int i = 0; i < inNumberFrames; i++)targetBuffer[i] = .5*sinf(i*rad);
    
    if(debug){
        int32_t sampleCount = availableBytes/sizeof(Float32);       
        printf("\n\tsampleCount = %d\n",sampleCount);
        printf("\n\tinNumberFrames = %d\n",inNumberFrames);
    }
    if(debug)printf("%d\n",gen.state);
    
    pthread_cond_signal (&cond);
    pthread_mutex_unlock (&mutex);
    
    return noErr;
}
//_____________________________________________________________________________
void resetToneUnit(void)
{
    for (int i = 0; i < partials; i++)
        *gen.phase = 0.;
}
//_____________________________________________________________________________
void play (int x, int y){
    wip();
    return;
}
//_____________________________________________________________________________
void stop (void){
    wip();
    return;
}
//_____________________________________________________________________________

#pragma mark - GL -
//_____________________________________________________________________________
int setup_glut(int* foo, char* bar[], int thewidth, int theheight)
{
    printf("Initializing GLUT\n");
    glutInit( foo, bar );
    //glut window
    glutInitDisplayMode( GLUT_RGBA | GLUT_ALPHA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL );
    glutInitWindowPosition (200, 50);
	glutInitWindowSize (thewidth, theheight);
	gMainWindow = glutCreateWindow(bar[0]);
    
    // read the data file, if any
    // compute all visualization parameters
    printf("        (overlay visuals %s\n\n",glutLayerGet(GLUT_OVERLAY_POSSIBLE)?"possible)":"not possible)");
    int i;
    glGetIntegerv(GL_MAX_CLIP_PLANES, &i);
    printf("max clip planes = %d\n", i);
    return gMainWindow;
}
//_____________________________________________________________________________
void init_gl (void)
{
	glEnable(GL_DEPTH_TEST);
    
	glShadeModel(GL_SMOOTH);
	glFrontFace(GL_CCW);
    
	glColor3f(1.0,1.0,1.0);
    
	glClearColor(0.0,0.0,0.0,0.0);         /* Background recColor */
	gCameraReset ();
	
	glPolygonOffset (1.0, 1.0);
	//SetLighting(4);
	//glEnable(GL_LIGHTING);
}
//_____________________________________________________________________________
void
gCameraReset(void)
{
    gCamera.aperture = 5;/*40*//*10*/
    gCamera.focalLength = 15;
    //orig
    gCamera.rotPoint = gOrigin;
    //our
    //gCamera.rotPoint = gRC;
    
    //orig
    /*
     gCamera.viewPos.x = 0.0;
     gCamera.viewPos.y = 0.0;
     */
    gCamera.viewPos.z = -gCamera.focalLength;
    
    //our
    gCamera.viewPos.x = 0.;
    gCamera.viewPos.y = 0.;
    
    
    
    gCamera.viewDir.x = -gCamera.viewPos.x;
    gCamera.viewDir.y = -gCamera.viewPos.y;
    gCamera.viewDir.z = -gCamera.viewPos.z;
    // orig
    /*
     gCamera.viewUp.x = 0;
     gCamera.viewUp.y = 1;
     */
    // our
    gCamera.viewUp.x = 0;
    gCamera.viewUp.y = 1;
    
    
    gCamera.viewUp.z = 0;
}
//_____________________________________________________________________________
void
drawGLString(GLfloat x, GLfloat y, char *string)
{
    int len, i;
    
    glRasterPos2f(x, y);
    len = (int) strlen(string);
    for (i = 0; i < len; i++) {
        glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, string[i]);
    }
}
//_____________________________________________________________________________
// text and synthesis snapshot
//_____________________________________________________________________________
void
drawGLText (GLint window_width, GLint window_height)
{
	char outString [256] = "";
	GLint matrixMode;
	GLint vp[4];
	GLint lineSpacing = 13;
	GLint line = 0;
	GLint startOffest = 7;
	
	glGetIntegerv(GL_VIEWPORT, vp);
	glViewport(0, 0, window_width, window_height);
	
	glGetIntegerv(GL_MATRIX_MODE, &matrixMode);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glScalef(2.0f / window_width, -2.0f / window_height, 1.0f);
	glTranslatef(-window_width / 2.0f, -window_height / 2.0f, 0.0f);
	
	// draw

	glDisable(GL_LIGHTING);
	glColor3f (.6, .5, .4);

	if (gShowInfo) {
        
		sprintf (outString, "Camera Position: (%0.1f, %0.1f, %0.1f)", gCamera.viewPos.x, gCamera.viewPos.y, gCamera.viewPos.z);
		drawGLString (10, window_height - (lineSpacing * line++) - startOffest, outString);

		sprintf (outString, "World Rotation: (%0.1f, %0.2f, %0.2f, %0.2f)", gWorldRotation[0], gWorldRotation[1], gWorldRotation[2], gWorldRotation[3]);
		drawGLString (10, window_height - (lineSpacing * line++) - startOffest, outString);

            sprintf (outString, flat ? "Projection: Flat" : "Projection: Perspective");
            drawGLString (10, window_height - (lineSpacing * line++) - startOffest, outString);

	}
	
	if (gShowHelp) {
		line = 1;
		sprintf (outString, "<shift> + L: translate; <opt> + L: rotate;\n");
		drawGLString (10, (lineSpacing * line++) + startOffest, outString);
		sprintf (outString, "<ctrl> + L: zoom; <opt> + M: synthesize;\n");
		drawGLString (10, (lineSpacing * line++) + startOffest, outString);
		sprintf (outString, "arrows: aperture & focal length\n");
		drawGLString (10, (lineSpacing * line++) + startOffest, outString);
		sprintf (outString, "H: toggle help\n");
		drawGLString (10, (lineSpacing * line++) + startOffest, outString);
		sprintf (outString, "I: toggle info\n");
		drawGLString (10, (lineSpacing * line++) + startOffest, outString);
        if(gen.midif)glColor3f (.7, .6, .2);
		sprintf (outString, "M: toggle MIDI\n");
		drawGLString (10, (lineSpacing * line++) + startOffest, outString);//multithreading, NF, freqDisp
        
        glColor3f (.6, .5, .4);
        sprintf (outString, "T: toggle Multithread\n");
        if(oldstuff && multithreading)glColor3f (.7, .6, .2);
        if((oldstuff == FALSE)&&(gen.wtable == FALSE))glColor3f (.6, .1, .2);
		drawGLString (10, (lineSpacing * line++) + startOffest, outString);
        glColor3f (.6, .5, .4);
        if(oldstuff && NF)glColor3f (.7, .6, .2);
        if((oldstuff == FALSE)&&(gen.wtable == TRUE))glColor3f (.6, .1, .2);
        sprintf (outString, "W: toggle Wavetable Multithread\n");
        drawGLString (10, (lineSpacing * line++) + startOffest, outString);
        glColor3f (.6, .5, .4);
        if(freqDisp)glColor3f (.7, .6, .2);
        sprintf (outString, "F: toggle Frequency display\n");
        drawGLString (10, (lineSpacing * line++) + startOffest, outString);
        if(oldstuff == FALSE){
        glColor3f (.6, .1, .2);
        sprintf (outString, "N: Full pthread \n");
        drawGLString (10, (lineSpacing * line++) + startOffest, outString);    
        }
       
	}
    
    //drawing synthesis snapshot
    if(debug)printf("gen.state = %d\n",gen.state);
    if((gen.state > 0)&&(gen.state < 3)){
        glBegin(GL_LINES);
        glColor3f(0.1, 0.45, 0.3);
        glVertex2i(Xpos,0);
        glVertex2i(Xpos,gCamera.screenHeight);
        glEnd();
    }
	//end drawing synthesis snapshot
    
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(matrixMode);
	
	glViewport(vp[0], vp[1], vp[2], vp[3]);
}


//_____________________________________________________________________________
void reshape (int w, int h)
{
    //without overlay support this is enough!
	glViewport(0,0,(GLsizei)w,(GLsizei)h);
	gCamera.screenWidth = w;
	gCamera.screenHeight = h;
	glutPostRedisplay();
}
//_____________________________________________________________________________
void draw_test_partials (void)
{
    //test shape
    int vert    = partials;
    int hor     = 64;
    GLfloat r, g, b;
    //static GLfloat textr, textg, textb;
    float z;
    float invert= 1./(float)(vert- 1);
    float invhor= 1./(float)(hor - 1);
    if(debug)printf("invhor = %f\n", invhor);
    
    //the scale!!!
    //z-axis
    float ampspan = -1.;
    if (gShowCoord){
    glBegin(GL_LINES);
    glColor3f (0.40, 0.32, 0.23);
    glVertex3f (0., 0., 0.);
        glVertex3f (0., 0., flat ? 0.: ampspan);
    glEnd();
    
    //debug z-axis
    glBegin(GL_LINES);
    glColor3f (0.40, 0.32, 0.23);
    glVertex3f (1., 0., 0.);
    glVertex3f (1., 0., flat ? 0.: ampspan);
    glEnd();
    
    //2nd debug z-axis
    glBegin(GL_LINES);
    glColor3f (0.40, 0.32, 0.23);
    glVertex3f (0., 1., 0.);
    glVertex3f (0., 1., flat ? 0.: ampspan);
    glEnd();
    
    /*vertical axis*/
    //_Drawing
    glBegin(GL_LINES);
    glColor3f (0.40, 0.32, 0.23);
    glVertex3f (0., 0., 0.);
    glVertex3f (0., 1., 0.);
    glEnd();
    //_endDrawing
    
    /*horizontal axis*/
    //_Drawing
    glBegin(GL_LINES);
    glVertex3f (0., 0., 0.);
    glVertex3f (1., .0, 0.);
    glEnd();
    //_endDrawing
    }
    // test partials
    for(int v = 0; v < vert ; v++){
        //z = -1.;
        z = (flat ? 0. : -zF);
        glColor3f(.5,0.05,.07);//point color
        glBegin(GL_POINTS);
        glVertex3f(0.05,invert*v, z); // the red start point
        glEnd();
        
        glBegin(GL_LINE_STRIP) ;
        glColor3f(0.,0.,0.);
        glVertex3f(0.05, invert*v, z); // @ red start point
        GLfloat finc = 0.5/(float)hor;
        GLfloat fs = invert*v;
        GLfloat f;
        for(int h = 0; h < hor; h++){
            r = .5*(hor-h)*invhor;
            g = .7*(h*invhor);
            b = .8;
            z = -0.05;
            f = (freqDisp ? fs*(1.+h*finc) : fs);
            glColor3f(r, g, b);//hor points
            glVertex3f(0.05 + invhor*h*xF*0.95, f, z);
            if(debug)if (h == 0)printf("y[%d] = %f\n",v, f);
            //glVertex3f(0.05, invert*v, z);
        }
        //z = -1.;
        //z = -zF;
        z = (flat ? 0. : -zF);
        glColor3f(0.,0.,0.);
        f = (freqDisp ? fs*(1.+(hor-1)*finc) : fs);
        glVertex3f(1.*xF, f, z);//end point
        glEnd();
        
        glColor3f(.5,0.05,.07);//point color
        glBegin(GL_POINTS);
        //glVertex3f(1.*xF,invert*v, z);
        glVertex3f(1.*xF, f, z);//red end point
        glEnd();
        
    }
    
}
//_____________________________________________________________________________
void draw_test_scene(recCamera gCamera)
{
 	//GLint viewport[4];
    
    glClear(GL_COLOR_BUFFER_BIT);
    //glPushMatrix();
    //by translating we move center of the scaling to object's center!
    glTranslatef(.5,.5,0.);
    glScalef(scal[0],scal[1],scal[2]); //new in GLUT version!
    glTranslatef(-.5,-.5,0.);
    
    draw_test_partials();
    // glPopMatrix();
    
}
//_____________________________________________________________________________
void maindisplay(void)
{
	GLdouble xmin, xmax, ymin, ymax;
	// far frustum plane
	GLdouble zFar = -gCamera.viewPos.z + gShapeSize * .5;
    //GLdouble zFar = 7.5;
    //zF = zFar*.5;
	// near frustum plane clamped at 1.0
	GLdouble zNear = MIN (-gCamera.viewPos.z - gShapeSize * .5, 1.);//our=.33; orig.=.5
	// window aspect ratio
	GLdouble aspect = gCamera.screenWidth / (GLdouble)gCamera.screenHeight;
    
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
    /**/
    //fixing aspect issue allows for distorting the object!!!
	if (aspect > 1.) {
		ymax = zNear * tan (gCamera.aperture * .5 * DTOR);
		ymin = -ymax;
		xmin = ymin /** aspect*/;
		xmax = ymax /** aspect*/;
	} else {
		xmax = zNear * tan (gCamera.aperture * .5 * DTOR);
		xmin = -xmax;
		ymin = xmin /*/ aspect*/;
		ymax = xmax /*/ aspect*/;
	}
    
    
    
	glFrustum(xmin, xmax, ymin, ymax, zNear, zFar*2.);
	
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
    //identity coordinates -1. to 1. for gluLookAt
    //orig
    /*
     gluLookAt (gCamera.viewPos.x, gCamera.viewPos.y, gCamera.viewPos.z,
     gCamera.viewPos.x  + gCamera.viewDir.x, //horizontal centering
     gCamera.viewPos.y + gCamera.viewDir.y , //vertical centering
     gCamera.viewPos.z + gCamera.viewDir.z,
     gCamera.viewUp.x, gCamera.viewUp.y ,gCamera.viewUp.z);
     
     */
    //our: sets eye coordinates and perspective center to screen center!
	gluLookAt (gCamera.viewPos.x - .5*fSize, gCamera.viewPos.y + .5*fSize, gCamera.viewPos.z, //eye coords.->centering!
               gCamera.viewPos.x  + gCamera.viewDir.x - .5*fSize, //horizontal centering
               gCamera.viewPos.y + gCamera.viewDir.y + .5*fSize, //vertical centering
               gCamera.viewPos.z + gCamera.viewDir.z,
               gCamera.viewUp.x, gCamera.viewUp.y ,gCamera.viewUp.z);
    
    //why do we need this?
	// track ball rotation
    //
    //glTranslated(-gCamera.screenWidth/2, -gCamera.screenHeight/2, 0);
    //glTranslatef(-.5*fSize, -.5*fSize, 0.);
	//glRotatef (gTrackBallRotation[0], gTrackBallRotation[1], gTrackBallRotation[2], gTrackBallRotation[3]);
    
    
	
    glTranslatef(trans[0], trans[1], trans[2]);
    glTranslatef(-.5,.5,0.);
    glRotatef(rot[0], 1.0, 0.0, 0.0);
    glRotatef(rot[1], 0.0, 1.0, 0.0);
    glTranslatef(.5,-.5,0.);
    
    //glTranslatef(.5*fSize, .5*fSize, 0.);
    
    //glTranslated(gCamera.screenWidth/2, gCamera.screenHeight/2, 0);
    //world rotation
	glRotatef (gWorldRotation[0], gWorldRotation[1], gWorldRotation[2], gWorldRotation[3]);
	
    //	glClearColor (0.2f, 0.2f, 0.4f, 1.0f);	// clear the surface
    glClearColor (0.f, 0.f, 0.1f, 1.0f);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    draw_test_scene(gCamera);
    //glPopMatrix();
	/**/
    //ShowPartials();
    
    // text and synthesis snapshot
	drawGLText (gCamera.screenWidth, gCamera.screenHeight);
    
	glutSwapBuffers();
}
//_________________________________________________________________
//_________________________________________________________________


void mouseTranslate (int x, int y)
{
	static GLfloat fx, fy;
	static int prev_x, prev_y, dx, dy;
	dx = x - prev_x;    //if( dx > 5)dx = 2; if( dx < -5 )dx = -2;
	dy = y - prev_y;	//if( dy < -5)dy = -2; if( dy > 5 )dy = 2;
	
	if(dragging){
        //        fx = dx*2./(GLfloat)(thewidth);
        //        fy = -dy*2./(GLfloat)(theheight);
        fx = dx*2./((GLfloat)(gCamera.screenWidth)*GHOST);
        fy = -dy*2./((GLfloat)(gCamera.screenHeight)*GHOST);
        //trans[0] += fx;//orig
        trans[0] -= fx; trans[1] += fy; trans[2]=0.;
		zero.x -= dx; zero.y += dy;
		one.x -= dx; one.y += dy;
        
        //	printf("X: %d\t Y: %d\t fX: %f\t fY: %f\n",x,y, fx, fy);
        //        draw_scene(gCamera);
        glutPostRedisplay();
	}
	prev_x = x; prev_y = y; dragging = GL_TRUE;
}
//_________________________________________________________________
//_________________________________________________________________

void mouseRot (int x, int y)
{
	/*This is important to understand: In GLUT
	 mouse callback functions can't be altered
	 by means of logical modifiers. Only the
	 final function it generates can!
	 */
	//GLfloat fz;
	static int prev_x, prev_y, dx, dy;
    if(!flat){
        if(dragging){
            dx = x - prev_x;    //if( dx > 5)dx = 2; if( dx < -5 )dx = -2;
            dy = y - prev_y;	//if( dy < -5)dy = -2; if( dy > 5 )dy = 2;
            
            rot[0] -= (dy * 180.)/500.; // orig +=
            rot[1] += (dx * 180.)/500.;
#define clamp(a) a = a > 360. ? a-360. : a < -360. ? a += 360. : a
            clamp(rot[0]);
            clamp(rot[1]);
            //draw_test_scene(gCamera);
            glutPostRedisplay();
        }
        prev_x = x; prev_y = y; dragging = GL_TRUE;
    }else{
        /*if(debug)*/printf("Flat\n");
    }
}
//__________________________________________________________________________________________//
void mouseScale (int x, int y)
{
	GLfloat fz;
	static float r, h_ratio, v_ratio;
	static int prev_x, prev_y, dx, dy;
	h_ratio = (float)gCamera.screenWidth/INITIAL_SIZE;
    v_ratio = (float)gCamera.screenHeight/INITIAL_SIZE;
	if(dragging){
        
        dx = x - prev_x; //if (dx > 2) dx = 1; if (dx < -2) dx = -1;
        dy = y - prev_y; //if (dy > 2) dy = 1; if (dy < -2) dy = -1;
        
        //	fz = (GLfloat)sqrtf((1 + (dx/(float)thewidth))*(1 + dx/(float)thewidth) + (1 + dy/(float)theheight)*(1 + dy/(float)theheight));
        
        fz = 2*(dx - dy)/(GLfloat)gCamera.screenWidth;
        //	fz = sqrt(x*x + y*y)/(GLfloat)thewidth;
        //	fz = x/(GLfloat)thewidth -.5;
		scal[0]= ((scal[0]>GHOST)||(fz>0.)?scal[0]+fz: scal[0]);
        scal[1]= ((scal[1]>GHOST)||(fz>0.)?scal[1]+fz: scal[1]);
		//scal[1]= scal[0]; //is this OK?
        scal[2]=1.;
		
		zoom = scal[0];
		r = zoom*sqrtf(h_ratio*h_ratio + v_ratio*v_ratio)/2.56f;
        //		zoom -= (dx+ .5*dy)/200.;
        //		printf("r = %f\n",r);
		/*define point size as far from the drawing loop as possible*/
		if(r >6.)ptsize = 4;
		else if(r >4.5)ptsize = 3;
		else if(r >3.)ptsize = 2;
		else ptsize = 1;
        
        //		trans[2] = zoom -1.;
        //	prev_x = x; prev_y = y;
        //	printf("X: %d\t fZ: %f\n",x,fz);
        //		final:
        //	draw_scene(gCamera);
        glutPostRedisplay();
	}
	prev_x = x; prev_y = y; dragging = GL_TRUE;
	return;
}
//_____________________________________________________________________________
void mouse (int button, int state, int x, int y)
{
    
    int modifier = glutGetModifiers();
    
    if ((button == GLUT_LEFT_BUTTON) && (state == GLUT_DOWN)){
        if (modifier == GLUT_ACTIVE_ALT){
            selecting = GL_FALSE;
            glutMotionFunc (mouseRot);
            glutPostRedisplay();
        }else if (modifier == GLUT_ACTIVE_SHIFT){
            selecting = GL_FALSE;
            glutMotionFunc (mouseTranslate);
            glutPostRedisplay();
        }
        else  if (modifier == GLUT_ACTIVE_CTRL){
            selecting = GL_FALSE;
            glutMotionFunc (mouseScale);
            glutPostRedisplay();
        }
        else;
    }else if((button == GLUT_LEFT_BUTTON) && (state == GLUT_UP)){
        if (modifier == GLUT_ACTIVE_ALT){
            glutMotionFunc (NULL);
            dragging = GL_FALSE;
            glutPostRedisplay();
        }else if (modifier == GLUT_ACTIVE_SHIFT){
            glutMotionFunc (NULL);
            dragging = GL_FALSE;
            glutPostRedisplay();
        }
        else  if (modifier == GLUT_ACTIVE_CTRL){
            glutMotionFunc (NULL);
            dragging = GL_FALSE;
            glutPostRedisplay();
        }
        else
            ;
    }
    
    
    if ((button == GLUT_MIDDLE_BUTTON) && (state == GLUT_DOWN)) {
        if (modifier == GLUT_ACTIVE_ALT){
            if(!(gen.midif)){
                glutMotionFunc(motion);
                //transformInput(); //converts x values to time values!
                //play(x,y);
                if(gen.state == idle){
                Xf = 1047.*x/(Float32)gCamera.screenWidth;
                gen.f[0] = updateFrequency(Xf/524.);
                for(int i = 1; i < partials; i++)
                    gen.f[i] = gen.f[0]*i;
                on = true;
                gen.state = starting;
                gen.playing = true;
                if(oldstuff)
                dproducer(on, multithreading);
                else{
                //create producer thread farm
                if(gen.wtable)
                    workThreadFarm = pdproducer(true, true, &gen);
                else
                    workThreadFarm = pdproducer(true, false, &gen);
                }
                Xpos = x;
                }
                glutPostRedisplay();
            }
        }else{
            ;
        }
	} else if ((button == GLUT_MIDDLE_BUTTON) && (state == GLUT_UP)){
        /*
        if (modifier == GLUT_ACTIVE_ALT) {
            //transformInput() converts x values to time values!
            if(!(gen.midif)){
                //stop();
                ;
            if(gen.state == active)gen.state = stopping;
                gen.playing = false;
            }else{
                if(gen.state == active) gen.state = stopping;
                gen.playing = false;
            }
            glutMotionFunc (NULL);
            glutPostRedisplay();
        }else*/
         { //no alt
            
            /*
            if (gen.playing == TRUE)
                stop();
             */
             if(gen.state == active){
                gen.state = stopping;
                gen.STOP = true;
             }
            gen.playing = false;
             if(!oldstuff){
            usleep(30000);
            //            mousePan (x, y);
            //gen.STOP = true;
            pthread_cancel(workThreadFarm);
             }
            glutMotionFunc (NULL);
            glutPostRedisplay();
        }
    }
    
}
//_____________________________________________________________________________
void key(unsigned char inkey, int px, int py)
{
    Boolean m = gen.midif;
    //int modifiers = glutGetModifiers();
    gLastKey = inkey;
    switch (inkey) {
            
        case '1':
        case '2':
        case '3':
        case '4':
            if (gen.state == idle)makeFreqEnv ((int)inkey-48, &gen);
            printf("%d\n", (int)inkey-48);
            break;
        case 27:
            dealloc_audio();
            pthread_mutex_destroy(&mutex);
            pthread_cond_destroy(&cond);
            printf("\tExiting program\n");
            exit(0);
            break;
        case 'c': // help
        case 'C':
            gShowCoord =  1 - gShowCoord;
            glutPostRedisplay();
            break;
            
        case 'h': // help
        case 'H':
            gShowHelp =  1 - gShowHelp;
            glutPostRedisplay();
            break;
        case 'i': // info
        case 'I':
            gShowInfo =  1 - gShowInfo;
            glutPostRedisplay();
            break;
        case 'm':
        case 'M':
            m = !m;
            midi = m;
            gen.midif = m;
            //printf("\nMIDIF = %d\n", (int)gen.midif);
            if(midi){
                //par = 1;
                //glutTimerFunc(FAST, timerCB, 1);
                //register
                if(tCallbackReg != GL_TRUE) {
                    tCallbackReg = GL_TRUE;
                    glutTimerFunc(FAST, timerCB, FAST);
                }

                printf("\nMIDI ON\n");

            }else{ //glutTimerFunc(FAST, timerCB, 0);//msecs = SLOW;
                printf("\nMIDI OFF\n");
            }
            glutPostRedisplay();
            break;
        case 'n':
        case 'N':
            if(gen.state != idle)
            break;
            oldstuff = !oldstuff;
            gen.wtable = !(gen.wtable);
            glutPostRedisplay();
            printf("Full Pthread: %s\n", oldstuff ? "OFF" : "ON");
            break;
            //toggle perspective
        case 'p':
        case 'P':
            flat = !flat;
            if (flat){
                rot[2] = rot[0];
				rot[3] = rot[1];
				rot[0] = rot[1] = 0.;
            }else{
                rot[0] = rot[2];
				rot[1] = rot[3];
				rot[2] = rot[3] = 0.;
            }
            glutPostRedisplay();
            /*
            if(fplayer.playing == FALSE){
                //synthesize buffer...
                //synthesizeBuffer(fBuffersize);
                fbplay0(fBuffersize);
            }
            else{
                fbstop0();
                //free buffer...
                //free(fBuffer);
            }
             */
            break;
        case 't':
        case 'T':
            if(oldstuff){
            if(gen.state == idle){
                if(NF) {
                 printf("command ignored\n");
                    break;
                }
                if(!NF)multithreading = !multithreading;
                resetToneUnit();
                printf("Multithreading: %s\n", (multithreading ? "ON": "OFF"));
                
            }
            }else gen.wtable = !gen.wtable;
            glutPostRedisplay();
            break;
        
        case 'w':
        case 'W':
            if(oldstuff){
            if(gen.state == idle){
                if(multithreading) {
                    printf("command ignored\n");
                    break;
                }
                if(!multithreading) NF = !NF;
                resetToneUnit();
                printf("MultiWavetable: %s\n", (NF ? "ON": "OFF"));
                
            }   
            } else gen.wtable = !gen.wtable;
            glutPostRedisplay(); 
            break;
         /**/
        case 'f':
        case 'F':
            freqDisp = !freqDisp;
            glutPostRedisplay();
            break;
        case '/':
            gShapeSize += .5;
            printf("gShapeSize = %f\n",gShapeSize);
            glutPostRedisplay();
            break;
        case '.':
            if(gShapeSize > 1.)
                gShapeSize -= .5;
            printf("gShapeSize = %f\n",gShapeSize);
            glutPostRedisplay();
            break;
        case ';':
            zF *= .9;
            if(zF < .5)zF = .5;
            glutPostRedisplay();
            break;
        case '\'':
            zF *= 1.1;
            if(zF > gShapeSize)zF = gShapeSize;
            glutPostRedisplay();
            break;
        case '[':
            //xF = (xF <= 1.1 ? 1. : xF*.909091);
            scal[0]= ((scal[0]>GHOST) ? scal[0]*.90909091: scal[0]);
            glutPostRedisplay();
            break;
        case ']':
            //xF = (xF >= 9.7 ? 10. : xF*1.1);
            scal[0]= ((scal[0]<10.) ? scal[0]*1.1: scal[0]);
            glutPostRedisplay();
        case '{':
            scal[1]= ((scal[1]>GHOST) ? scal[1]*.90909091: scal[1]);
            glutPostRedisplay();
            break;
        case '}':
            scal[1]= ((scal[1]<10.) ? scal[1]*1.1: scal[1]);
            glutPostRedisplay();
            break;
        default:
            break;
            
    }
}

//_____________________________________________________________________________
void m2xy( int m, int* X, int* Y)
{
    //float ex, why;
    
    return;
}
//_____________________________________________________________________________
/*
 SJ:
 This motion function only reports pointer position and updates the number of
 candidate partials for the puropse of AdditiveRenderProc() and cursor redraw
 */
//_____________________________________________________________________________
void motion (int x, int y)
{    
    //transformInput() converts x values to time values! (not used here!)
    //this primitive algorithm maps mouse coordinates into Xf-Yc/Yf values.
    //these values dynamically control fundamental frequency and number of partials
    if(0)printf("[%d]\t[%d]\n", x, y);
    if( x <= 0 ){
        Xpos = 0;
        Xf = 0.;
        Yf = 0.;
        glutPostRedisplay();
        return;
    }
    else if( x >= gCamera.screenWidth ){
        Xpos = gCamera.screenWidth -1;
        Xf = 1047.;
        Yf = 0.;
        glutPostRedisplay();
        return;
    }
    else {
        Xpos = x;
        Xf = 1047.*x/(Float32)gCamera.screenWidth;
        Yf = (gCamera.screenHeight - y)/(Float32)gCamera.screenHeight;
    }
    
    //y coordinates do downwards...
    if( y <= 0){
        Yf = 0.;
        Yc = 0.;
    }
    else if( y >= gCamera.screenHeight) {
        Yf = 0.;
        Yc = 0.;
        glutPostRedisplay();
    }
    else{
        Yf = (gCamera.screenHeight - y)/(Float32)gCamera.screenHeight;
        Yc = Yf;
    }

    //update number of candidates according to Yc mouse position!!!
    
    if(0)printf("Height: %d\ty = %d\tYc= %f\tYf = %f\n", gCamera.screenHeight, y, Yc,Yf);
    if(0)printf("Width: %d\tx = %d\tXf = %f\n", gCamera.screenWidth, x, Xf);
    candidates = (UInt32)(partials*Yc);
    //if((Xf > 1000.)||(Xf < 47.))candidates = 0;
    gen.candidates = candidates;
    
    //numCand = candidates;
    if(debug)printf("Candidates: %d\n",candidates);
    gen.f[0] = updateFrequency(Xf/524.);
    if(debug)printf("gen.f = %f\n", gen.f[0]);
    for(int i = 1; i < partials; i++)
        gen.f[i] = gen.f[0]*i;
    
    glutPostRedisplay();
}
//_____________________________________________________________________________

//_____________________________________________________________________________
void special(int key, int px, int py)
{
    gLastKey = key;
    switch (key) {
        case GLUT_KEY_UP: // arrow forward, close in on world
            break;
            gCamera.focalLength -= .5f;
            if (gCamera.focalLength < 0.f)
                gCamera.focalLength = 0.f;
            glutPostRedisplay();
            break;
        case GLUT_KEY_DOWN: // arrow back, back away from world
            break;
            gCamera.focalLength += .5f;
            glutPostRedisplay();
            break;
        case GLUT_KEY_RIGHT: // arrow left, smaller aperture
            break;
            gCamera.aperture -= .5f;
            if (gCamera.aperture < .5f) //our = .5f; orig = 0.f
                gCamera.aperture = .5f;
            glutPostRedisplay();
            break;
        case GLUT_KEY_LEFT: // arrow right, larger aperture
            break;
            gCamera.aperture += .5f;
            //our: 35.f max check added
            if (gCamera.aperture > 10.f)
                gCamera.aperture = 10.f;
            glutPostRedisplay();
            break;
    }
}

//_____________________________________________________________________________
//_____________________________________________________________________________

int main(int argc, const char * argv[])
{

    char text[16];
    debug = false;
    partials = 16;
    numCand  = 4;
    f0 = 130.75;
    
    int gMainWindow = 0;
    GLint thewidth;
    GLint theheight;
    thewidth = theheight = INITIAL_SIZE;
    scal[2]=scal[1]=scal[0]=GHOST; zoom = GHOST;//it's a hack!!!
    rot[3]=rot[2]=rot[1]=rot[0]=0.;
    
    fBuffersize = 88200;
    
    if (argc == 3){
    float p = (int)(*argv[1]);    
    f0 = (float)(*argv[2]);
    partials = p > MAXP ? MAXP : p;    
    }
    printf("GLAS - threaded  (additive)  synthesizer v0.88\n");
    printf("     - using POSIX threads, mutex & conditions\n");
    printf("     - with OpenGL visualization and interface\n");
    printf("     - with monophonic MIDI controller");
    printf("______________________________________________\n\n");
    if (argc == 3){
        int p = atoi(argv[1]);
        f0 = (float)atof(argv[2]);
        partials = p > MAXP ? MAXP : p;
    }else{
    printf("Enter the number of partials: (1-%d):\n\t> ", MAXP);
    scanf("%d", &partials);
    if(partials <   1)  partials =   1;
    if(partials > MAXP) partials = MAXP;
        

    printf("Enter fundamental frequency:\n\t> ");
    scanf("%s", text);
    if(atoi(text)==!atoi(text)){
        printf("using default fundamental C=130.75Hz\n");
    }else f0 = atoi(text);
    }
    printf("Generating %d harmonics of %f Hz fundamental\n",partials, f0);
    struct timespec ts;
    ts.tv_sec  = 0;
    ts.tv_nsec = 30000000;
    debug = false;
    gen.partials = partials;
    gen.numCand  = gen.partials;
    gen.cpuNum   = getCPUnum();
    //create the mutex and cond to coordinate between producer and consumer threads
    pthread_cond_init  (&cond, NULL);
    pthread_mutex_init (&mutex, NULL);
    
    //setup audio
    setup_audio();
    
    //setup GLUT
    int foo = 1; char * bar[] = { "GLAS" };
    
    gMainWindow = setup_glut(&foo, bar, thewidth, theheight);

    init_gl(); // standard GL init - function which draws everything!!!
    
    glutReshapeFunc (reshape);
    glutDisplayFunc (maindisplay); //which calls draw_scene(), which calls show_partials()
	glutKeyboardFunc (key);
	glutSpecialFunc (special);
	glutMouseFunc (mouse);
    glutMotionFunc(motion);
    msecs = FAST;
    //glutTimerFunc(msecs, timerCB, msecs);
    tCallbackReg = GL_FALSE;
    par = 0;
    //glutTimerFunc(FAST, update, 0);
    
    flat = 0;
    zF = -gCamera.viewPos.z*.1;// + gShapeSize ;
    xF = 1.;
    
    
    //here comes menu function, if any
    
    //initialize navigation variables
    
    // GLUT main loop
    glutMainLoop();
    

escape_cleanup:
    dispatch_release(mSemaphore);
    pthread_mutex_destroy(&mutex);
    pthread_cond_destroy(&cond);
	AudioOutputUnitStop(gen.toneUnit);
	AudioUnitUninitialize(gen.toneUnit);
	AudioComponentInstanceDispose(gen.toneUnit);
    if(pthread_cancel(workThreadFarm) == 0)printf("   ...success\n");
    printf("\nTP:  execution finished - Audio SHUT DOWN\n");
    printf("_____________________________________________\n\n");
    return 0;
}

