//
//  glas.h
//  GLAS
//
//  Created by Stanko Juzbasic on 12/19/15.
//  Copyright (c) 2015 Stanko Juzbasic. All rights reserved.
//

#ifndef TP_tp_h
#define TP_tp_h
#import "TPCircularBuffer.h"
#import "cbuffer.h"
#import <AudioToolbox/AudioToolbox.h>
#import <AudioUnit/AudioUnit.h>
#import <dispatch/dispatch.h>

#include <sys/sysctl.h>
#include <pthread.h>
#include <GLUT/glut.h>
#include <OpenGL/glext.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>

#define idle 0
#define active 1
#define starting 2
#define stopping 3
#define stopped 4

#define MAXP 4096

#define DTOR 0.0174532925
#define CROSSPROD(p1,p2,p3) \
p3.x = p1.y*p2.z - p1.z*p2.y; \
p3.y = p1.z*p2.x - p1.x*p2.z; \
p3.z = p1.x*p2.y - p1.y*p2.x

#define NYQUISTF 20000.

#define fSize 1.
#define INITIAL_SIZE 450.0
#define GHOST 1.0f

#define FAST 100
#define SLOW 10000

// Declare workThreadFarm, cond and mutex to be available to multiple threads
pthread_t workThreadFarm;
pthread_cond_t cond;
pthread_mutex_t mutex;
//

Boolean debug;
Boolean multithreading;
Boolean NF;




// Declare mSemaphore somewhere it is available to multiple threads
dispatch_semaphore_t mSemaphore;

Float64 R;
int playing;
UInt32 frames;
int32_t availableBytes;
Float32 **fbuffW;

Boolean debug;
UInt32 numCand;
Float32 f0;

//work-in-progress audio controller
typedef struct {
    TPCircularBuffer            buffer;
    UInt32                      fillThreshold;
    cbuf_t                      cbuffer;
    UInt32                      partials;
    UInt32                      candidates;
    UInt32                      numCand;
    AudioComponentInstance      toneUnit;
    Float64                     sampleRate;
    AudioStreamBasicDescription streamFormat;    
    Float32*                     f;     //frequency
    Float32*                     a;     //amp.
    Float32*                     prevf; //prev. frequency
    Float32*                     preva; //prev. amp.
    Float32*                     phase;
    Float32*                     previous_phase_increment;
    Float32*                     val;
        int*                    arg;
        int*                    previous_arg;
    UInt32                      frames;
    int                         state;
    Boolean                     midif; //wip
    Boolean                     playing; //obsolete
    double                      startingFrameCount;
    float                       time;
    //
    MIDIClientRef               midiclient;
    MIDIPortRef                 midiin;
    Boolean                     STOP;
    int                         cpuNum;// number of logical CPUs
    Boolean                     on;    //a control flag
    Boolean                     wtable;
    int                         freqEnv;
} wipAudioController; //in RtAudio -> uData

wipAudioController gen;

int    nThreads;

typedef struct data{
    int tid;
    int cpuCount;
}data;

typedef struct tData{
    int tid;
    int cpuCount;
    wipAudioController* genD;
}tData;

//MIDI information
Boolean midi;
Boolean omni;
int     hold; //tracks how many notes have been held
int     prev_hold;
Byte    n[16];//array of held notes
Byte    channel;//MIDI active channel number
Float32 pitchf;//read from MIDI

Float32 mmif; //midi modulation index (float) (0.-1.)
Float32 mxpf; //midi expression CC11  (float) (1.-4.)
Byte    bmi;  //midi byte debug value
Boolean changed;//modulation depth changed
Boolean changex;//expression ctrl. changed
Float32 bendf; //bend value (float)
Float32 benddf;//bend depth (float)

//screen coordinates in audio space
GLboolean tCallbackReg;
Float32 Xf, Yf, Yc;

UInt32      partials;
UInt32      candidates;
Float32*    Sin;

Byte    n[16];//array of held notes
Float32 amplitudef; //junk
long fBuffersize;   //junk;

void setup_audio (void);
void dealloc_audio (void);
void createToneUnit(wipAudioController *gen);

static int checkResultLite(OSStatus error, const char *operation);

int  getCPUnum(void);
void resetToneUnit(void);
Float32* makeSin (int);

void producer  (Boolean on);
void sproducer  (Boolean on);
void dproducer (Boolean on, Boolean multithreading);

void producerSum  (Boolean on);
void *producerT   (void *T);
void *producerTN  (void *T);
Float32 updateFrequency(float sliderValue);

static OSStatus audioRenderCallback(void *inRefCon,
                                    AudioUnitRenderActionFlags *ioActionFlags,
                                    const AudioTimeStamp *inTimeStamp,
                                    UInt32 inBusNumber,
                                    UInt32 inNumberFrames,
                                    AudioBufferList *ioData);

void midiReadProc(const MIDIPacketList *packetList,
                  void* readProcRefCon,
                  void* srcConnRefCon);

//Float32 **fbuffW;
void play (int, int);
void stop (void);

int fbplay0 (long size);
int fbstop0 (void);
void wip(void);
Boolean on;

//_____________________________________________________________
#pragma mark - GL globals -
//_____________________________________________________________
GLdouble zF;
GLdouble xF;
enum {
	kTextureSize = 256
};

GLint Xpos;

typedef struct {
    GLfloat x,y,z;
} recVec;

typedef struct {
	recVec viewPos; // View position
	recVec viewDir; // View direction vector
	recVec viewUp; // View up direction
	recVec rotPoint; // Point to rotate about
	GLdouble focalLength; // Focal Length along view direction
	GLdouble aperture; // gCamera aperture
	GLdouble eyeSep; // Eye separation
	GLint screenWidth,screenHeight; // current window/screen height and width
} recCamera;

typedef struct xPoint{
	GLdouble x; GLdouble y; GLdouble z;
} xpoint;

GLint ptsize;

typedef struct {
	float xt;
	float yt;
	float zt;
} test;

test tuplet;
static float zoom;

//what do we need this for:
//GLfloat gShapeSize = 11.0f;
GLfloat gShapeSize = 5.f;
//orig
//GLint gDollyPanStartPoint[2] = {0, 0};
//our
GLint gDollyPanStartPoint[2] = {200, 200};
//orig
GLfloat gTrackBallRotation [4] = {0.0, 0.0, 0.0, 0.0};
//our
//GLfloat gTrackBallRotation [4] = {0.0, 0.5, 0.5, 0.0};
GLboolean gDolly = GL_FALSE;
GLboolean gPan = GL_FALSE;
GLboolean gTrackBall = GL_FALSE;
//orig:
//GLfloat gWorldRotation [4] = {155.0, 0.0, -1.0, 0.0};
//our:
GLfloat gWorldRotation [4] = {180.0, 0.0, -1.0, 0.0};

GLboolean gLines = GL_FALSE;
GLboolean gPolygons = GL_TRUE;

GLboolean gShowHelp = GL_TRUE;
GLboolean gShowInfo = GL_TRUE;
GLboolean gShowCoord= GL_TRUE;
recCamera gCamera;

//
recVec gOrigin = {0., 0., 0.};
recVec gRC = {.5,.5,0.};
//
int gLastKey = ' ';

int gMainWindow = 0;

GLfloat trans[3];	/* current translation */
GLfloat rot[4];		/* current rotation 0,1; reset rotation cache 2,3*/
GLfloat scal[3];

//not needed
/*
 GLuint gPointList = (GLuint)NULL;
 GLuint gWireList = (GLuint)NULL;
 GLuint gSolidList = (GLuint)NULL;
 */
static GLboolean dragging = GL_FALSE; //makes sure that we can start dragging from anywhere in the screen!
static GLboolean selecting = GL_FALSE;
//int flat;
xpoint zero, one;
static GLboolean flat = GL_FALSE;

static GLboolean freqDisp = GL_FALSE;

//_____________________________________________________________
#pragma mark - GL prototypes -
//_____________________________________________________________
void gCameraReset(void);

void
drawGLString(GLfloat x, GLfloat y, char *string);

//void SetLighting(unsigned int mode);

void drawGLText (GLint window_width, GLint window_height);

void init (void);

void reshape (int w, int h);

void draw_test_partials (void);

void draw_test_scene(recCamera gCamera);

void maindisplay(void);

void special(int key, int px, int py);

//void mouseDolly (int x, int y);

//void mousePan (int x, int y);

//void mouseTrackball (int x, int y);

void mouseRot (int x, int y);

void mouseTranslate (int x, int y);

void mouseScale (int x, int y);

void mouse (int button, int state, int x, int y);

void motion (int, int);

void key(unsigned char inkey, int px, int py);

//void spaceballmotion (int x, int y, int z);

//void spaceballrotate (int rx, int ry, int rz);




#endif
